#######################################################################
# Copyright (c) 2023 Josh Luthy BD Life Sciences - Informatics,
# Ashland, OR
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################
## Need EmbedSOM >= 2.1.1

## For debugging to see what FlowJo is passing to the script.
##
## cat("FJ-PAR-PREFIX                = FJ_PAR_PREFIX \n")
## cat("FJ-PAR-DER-PAR-NAME          = FJ_PAR_DER_PAR_NAME \n")
## cat("FJ-SOM-FILE-PATH             = FJ_SOM_FILE_PATH \n")
## cat("FJ-DATA-FILE-PATH            = FJ_DATA_FILE_PATH \n")
## cat("FJ-CSV-OUPUT-FILE            = FJ_CSV_OUPUT_FILE \n")
## cat("FJ-PARAMS-LIST               = FJ_PARAMS_LIST \n")
## cat("FJ-PAR-PREFIX                = FJ_PAR_PREFIX \n")
## cat("FJ-PAR-SMOOTH                = FJ_PAR_SMOOTH \n")
## cat("FJ-PAR-K                     = FJ_PAR_K \n")
## cat("FJ-PAR-ADJUST                = FJ_PAR_ADJUST \n")
## cat("FJ-PAR-ADD-CELLIDS-TO-RESULT = FJ_PAR_ADD_CELLIDS_TO_RESULT \n")
## cat("FJ-PAR-SEED                  = FJ_PAR_SEED \n")
## cat("FJ-PAR-UMAP-DIST             = FJ_PAR_UMAP_DIST \n")


parNames <- c(FJ_PARAMS_LIST)
supLabel <- "FJ_SUPER_LABEL"
supLabelbkup <- "FJ_SUP_LABEL_BKUP"
addCellIdsToResult <- FJ_PAR_ADD_CELLIDS_TO_RESULT
EventNumberDP <- "EventNumberDP"
CellIdDP <- "CellId"
fjPrefix <- "FJ_PAR_PREFIX"
fjDerParName <- "FJ_PAR_DER_PAR_NAME"
fjSomFilePath <- "FJ_SOM_FILE_PATH"
fjDataFilePath <- "FJ_DATA_FILE_PATH"
fjCsvOutputFile <- "FJ_CSV_OUPUT_FILE"
outputFolder <- "FJ_OUTPUT_DIRECTORY_PATH"
fjPopName <- "FJ_POPNAME"
fjDims <- 2
fjSmooth <- FJ_PAR_SMOOTH # good range -3 (sharper) to 3 (smoother)
fjK <- FJ_PAR_K # good range 5 to 100
fjAdjust <- FJ_PAR_ADJUST / 100.0 # good range 0 to 3
fjSeed <- FJ_PAR_SEED # any
fjUmap.Dist <- FJ_PAR_UMAP_DIST # good range 0.1 to 2
fjXdim <- FJ_PAR_XDIM # it's sometimes useful to asymmetric SOM (Xdim != Ydim) between 5 to 64
fjYdim <- FJ_PAR_YDIM # it's sometimes useful to asymmetric SOM (Xdim != Ydim) between 5 to 64
fjRLEN <- 15 # good range here is between 5 to 50
fjMapSize <- FJ_PAR_MAPSIZE # 1000 is good in almost all cases. Good range 50 to 5000
fjSuperWeight <- FJ_PAR_WEIGHT #values can range from 0.0 to 1.0
fjEmbedSom.Map <- "FJ_PAR_EMBEDSOM_MAP"
fjFlowSOM.Map <- "FJ_PAR_FLOWSOM_MAP"
fjUMAP.Map <- "FJ_PAR_UMAP_MAP"
fjTSNE.Map <- "FJ_PAR_TSNE_MAP"
fjSuper.Map <- "FJ_PAR_SUPER_MAP"

if (supLabel == "fsupervisedID") {
	supLabel <- supLabelbkup
}

if (fjDerParName == "") {
	if (fjEmbedSom.Map) {
		derParName <- paste0("ES.", fjPopName, ".")
	} else if (fjFlowSOM.Map){
		derParName <- paste0("ES.fSOM.", fjPopName, ".")
	} else if (fjUMAP.Map) {
		derParName <- paste0("ES.UMAP.", fjPopName, ".")
	} else if (fjTSNE.Map) {
		derParName <- paste0("ES.tSNE.", fjPopName, ".")
	} else if (fjSuper.Map){
		derParName <- paste0("ES.sUMAP.", fjPopName, ".")
	} else {
		derParName <- paste0("EmbedSOM.ID.", fjPopName, ".")
	}
}

#    else {
# 	derParName <- paste0("EmbedSOM_ID_", fjDerParName, "_")
# }

stopAndReturnError <- function(message, ...) {
	cat(message, file=paste0(fjCsvOutputFile, "_err.txt"))
	stop(message, ...)
}

#######################################################################
###### Start package import or installations
#######################################################################
# Install function for packages
loadPackages <-function(...) {
	libs<-unlist(list(...))
	req<-unlist(lapply(libs,require,character.only=TRUE))
	need<-libs[req==FALSE]
	if(length(need)>0){
		install.packages(need, repos="https://cran.r-project.org")
		lapply(need,require,character.only=TRUE)
	}
}

loadPackages(c('ggplot2', 'FNN', 'igraph', 'Matrix', 'Rtsne', 'umap', 'uwot', 'utils', 'devtools', 'data.table', 'dplyr'))

if(!require("EmbedSOM",character.only=TRUE)){
	suppressMessages(suppressWarnings(devtools::install_github('exaexa/EmbedSOM')))
}

tryCatch({
	library("EmbedSOM")
}, error = function(e) {
	message <- "EmbedSOM was not able to run. There is an error loading the EmbedSOM R package. Please see How-To PDF and try to manually install or contact flowjo@bd.com."
	StopAndReturnError(message = message)
})

sessionInfo()

ev <- as.character(packageVersion("EmbedSOM"))
if(compareVersion(ev, "2.1.1") < 0 || compareVersion(ev, "3") >=0)
	stopAndReturnError("Incompatible EmbedSOM version")

## Returns a list of 3 elements
## X = a matrix of data from a CSV file as exported by FlowJo
## cellIds = a matrix with CellIds (if available)
## eventNumbers = a matrix with EventNumberDP (if available)
##
## csvPath: path the the CSV file as exported from FlowJo (including header with both stain and parameter is recommended when exporting from FlowJo)
readFlowJoCSV <- function(csvPath) {
	data <- data.table::fread(csvPath, check.names=FALSE)
	data <- as.data.frame(data)
	cellIdsColumn <- rep(NA, nrow(data))
	cellIds <- data.frame(CellIdDP = cellIdsColumn)
	names(cellIds) <- CellIdDP
	if (CellIdDP %in% names(data)) {
		cellIds <- data[, CellIdDP]
		data <- data[, -which(names(data) %in% c(CellIdDP))]
	}
	
	EventNumberColumn <- rep(NA, nrow(data))
	eventNumbers <- data.frame(EventNumberDP = EventNumberColumn)
	names(eventNumbers) <- EventNumberDP
	if (EventNumberDP %in% names(data)) {
		eventNumbers <- data[, EventNumberDP]
		data <- data[, -which(names(data) %in% c(EventNumberDP))]
	}
	
	# Trim the leading and tailing whitespaces
	colnames(data) <- trimws(colnames(data), "both")
	list(X=as.matrix(data), cellIds=as.matrix(cellIds), eventNumbers=as.matrix(eventNumbers))	
}

tmpRet = readFlowJoCSV(fjDataFilePath);
## inputMatrix <- tmpRet$X ## We don't need the data, just the cell ids or event numbers
cellIds <- tmpRet$cellIds
eventNumbers <- tmpRet$eventNumbers
colnames(eventNumbers) <- c(EventNumberDP)

if(fjFlowSOM.Map) {
	previousOBJ <- readRDS(fjSomFilePath)
	fSOM <- previousOBJ@FlowSOMClusters$som
	# clean up parameter names from the data matrix to match FlowSOM
	myColNames <- gsub( "\\::.*" , "", colnames(tmpRet$X), ignore.case = TRUE )
	myColNames <- trimws(myColNames, "r") # remove trailing whitespace
	colnames(tmpRet$X) <- myColNames # add the cleaned names back to data matrix

	outputMatrix <- EmbedSOM::EmbedSOM(data = tmpRet$X, fsom = fSOM, smooth = fjSmooth, k = fjK, adjust = fjAdjust, parallel = T)

	if ((NA %in% outputMatrix) || (NaN %in% outputMatrix)) {
		## Scaling can sometimes help
		fSOM$data <- scale(fSOM$data)
		outputMatrix <- EmbedSOM::EmbedSOM(data = tmpRet$X, fsom=fSOM, smooth = fjSmooth, k = fjK, adjust = fjAdjust, parallel = T)
	}

	if ((NA %in% outputMatrix) || (NaN %in% outputMatrix)) {
		stopAndReturnError("Unfortunately, we were unable to calculate results for this input\n(NaN in the resulting matrix). You may want to change the input arguments.")
	} else {
		if (addCellIdsToResult && !(NA %in% cellIds)) {
			## SeqGeq
			outputMatrix <- cbind(cellIds, outputMatrix)
			colnames(outputMatrix) <- c(CellIdDP, paste0(derParName, 1:fjDims, ".", fjPrefix))
			## Header to specify that those are metadata parameters, not genes.
			## That should work with CSVSample.mergeDataFile in SeqGeq, but it does not seem to work reliably, and it does not seem to
			## recognize derived parameters as metadata anyway, so let's just do it without a header for now
			##
			## cat("[Metadata]\n$Organization\tParamsInCols\n$ParamTypes\tIm\n$EventTypes\tIv\n\n[Data]\n", file = fjCsvOutputFile)
			## write.table(outputMatrix, file=fjCsvOutputFile, row.names=FALSE, quote=FALSE, append=TRUE, sep=",")
			write.csv(outputMatrix, file=fjCsvOutputFile, row.names=FALSE, quote=FALSE)
		} else {
			## FlowJo
			colnames(outputMatrix) <- paste0(derParName, 1:fjDims,".", fjPrefix)
			write.csv(cbind(eventNumbers, outputMatrix), file=fjCsvOutputFile, row.names=FALSE, quote=FALSE)
		}
	}
} else {
	set.seed(fjSeed) # all 3 map types below require random numbe
	if(fjEmbedSom.Map){
		map <- EmbedSOM::SOM(tmpRet$X, xdim=fjXdim, ydim=fjYdim, parallel=T, batch=T, rlen=fjRLEN)
	}
	if(fjTSNE.Map){
		map <- EmbedSOM::RandomMap(tmpRet$X, min(fjMapSize, nrow(data)), coordsFn=EmbedSOM::tSNECoords())
	}
	if(fjUMAP.Map) {
		map <- EmbedSOM::RandomMap(tmpRet$X, min(fjMapSize, nrow(data)), coordsFn=EmbedSOM::uwotCoords(min_dist=fjUmap.Dist))
	}
	if(fjSuper.Map){
		tempMat <- as.data.frame(tmpRet$X)
		clusterMat <- as.matrix(tempMat[supLabel])
		clusterMat <- dplyr::na_if(clusterMat, 0)
		clusterMat <- as.factor(clusterMat)
		print(levels(clusterMat))
		map <- EmbedSOM::RandomMap(tempMat, min(fjMapSize, nrow(data)), coordsFn=EmbedSOM::uwotCoords(uwotFn = uwot::umap,
																									   min_dist = fjUmap.Dist,
																									   y = clusterMat,
																									   target_weight = fjSuperWeight))
	}


	embed <- EmbedSOM::EmbedSOM(data=tmpRet$X, map=map, parallel=T, k=fjK, adjust=fjAdjust, smooth=fjSmooth)
	# embed is ready to be sent back to FlowJo.

	if (addCellIdsToResult && !(NA %in% cellIds)) {
		## SeqGeq
		embed <- cbind(cellIds, embed)
		colnames(embed) <- c(CellIdDP, paste0(derParName, 1:fjDims,".", fjPrefix))
		## Header to specify that those are metadata parameters, not genes.
		## That should work with CSVSample.mergeDataFile in SeqGeq, but it does not seem to work reliably, and it does not seem to
		## recognize derived parameters as metadata anyway, so let's just do it without a header for now
		##
		## cat("[Metadata]\n$Organization\tParamsInCols\n$ParamTypes\tIm\n$EventTypes\tIv\n\n[Data]\n", file = fjCsvOutputFile)
		## write.table(outputMatrix, file=fjCsvOutputFile, row.names=FALSE, quote=FALSE, append=TRUE, sep=",")
		write.csv(embed, file=fjCsvOutputFile, row.names=FALSE, quote=FALSE)
	} else {
		## FlowJo
		colnames(embed) <- paste0(derParName, 1:fjDims,".", fjPrefix)
		write.csv(cbind(eventNumbers, embed), file=fjCsvOutputFile, row.names=FALSE, quote=FALSE)
	}
}

message <- c("Run Settings\n",
"Run ID:", fjPrefix, "\n",
"Selected Population:", fjPopName, "\n",
"Smoothing:", fjSmooth, "\n",
"K value:", fjK, "\n",
"Adjust value:", fjAdjust, "\n",
"SOM X Grid Size:", fjXdim, "\n",
"SOM Y Grid Size:", fjYdim, "\n",
"Seed value:", fjSeed, "\n",
"UMAP distance value:", fjUmap.Dist, "\n",
"Map Size value:", fjMapSize, "\n",
"Run Supervised UMAP:", fjSuper.Map, "\n",
"Supervised ID Label:", supLabel, "\n",
"Backup Supervised ID:", supLabelbkup, "\n",
"Supervised Weighting:", fjSuperWeight, "\n",
"Run EmbedSOM:", fjEmbedSom.Map, "\n",
"Run FlowSOM Embedding:", fjFlowSOM.Map, "\n",
"Run tSNE:", fjTSNE.Map, "\n",
"Run Standard UMAP:", fjUMAP.Map, "\n")

cat(message, file=paste0(outputFolder, "RunSettings_", fjPrefix, "_.txt"))

# R seems to be saving the .RData when exiting, so let's clean up to at least make that tiny (i.e., empty environment)
rm(list=ls())
